//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "parallel/PetscSolver.h"


namespace AMDiS {

  using namespace std;

  PetscErrorCode myKSPMonitor(KSP ksp, PetscInt iter, PetscReal rnorm, void *)
  {    
    if (iter % 100 == 0 && MPI::COMM_WORLD.Get_rank() == 0)
      cout << "[0]  Petsc-Iteration " << iter << ": " << rnorm << endl;

    return 0;
  }


  PetscSolver* PetscSolver::getPetscSolver(Mat& petscMatrix, MeshDistributor *dist, int n)
  {
    FUNCNAME("PetscSolver::getPetscSolver()");

    string name("");
    GET_PARAMETER(0, "parallel->solver", &name);
    
    if (name == "petsc-schur") {
      return new PetscSolverSchur(petscMatrix, dist, n);
    } else if (name == "petsc" || name == "") {
      return new PetscSolverGlobalMatrix(petscMatrix, dist, n);
    } else {
      ERROR_EXIT("No parallel solver %s available!\n", name.c_str());
    }
  }


  void PetscSolverGlobalMatrix::providePetscSolver(KSP &solver, PC &pc)
  {
    FUNCNAME("PetscSolverGlobalMatrix::providePetscProblemStat()");
    
    KSPCreate(PETSC_COMM_WORLD, &solver);
    KSPGetPC(solver, &pc);
    
    KSPSetOperators(solver, matrix, matrix, SAME_NONZERO_PATTERN); 
    KSPSetTolerances(solver, 0.0, 1e-8, PETSC_DEFAULT, PETSC_DEFAULT);
    KSPSetType(solver, KSPBCGS);
    KSPMonitorSet(solver, myKSPMonitor, PETSC_NULL, 0);
    KSPSetFromOptions(solver);
    PCSetFromOptions(pc);
  }


  void PetscSolverSchur::providePetscSolver(KSP &solver, PC &pc)
  {
    FUNCNAME("PetscSolverSchur::providePetscProblemStat()");
    
    typedef map<int, DofContainer> RankToDofContainer;
    typedef map<DegreeOfFreedom, bool> DofIndexToBool;
    
    std::set<DegreeOfFreedom> boundaryDofsSet;
    std::set<DegreeOfFreedom> boundaryLocalDofs;
    RankToDofContainer& sendDofs = meshDistributor->getSendDofs();
    for (RankToDofContainer::iterator rankIt = sendDofs.begin(); 
	 rankIt != sendDofs.end(); ++rankIt)
      for (DofContainer::iterator dofIt = rankIt->second.begin(); 
	   dofIt != rankIt->second.end(); ++dofIt) {
	boundaryLocalDofs.insert(**dofIt);
	  
	for (int i = 0; i < nComponents; i++)
	  boundaryDofsSet.insert(meshDistributor->mapLocalToGlobal(**dofIt) * nComponents + i);
      }
      
    vector<DegreeOfFreedom> boundaryDofs(boundaryDofsSet.begin(), 
					 boundaryDofsSet.end());
      
    std::set<DegreeOfFreedom> otherBoundaryLocalDofs;
    RankToDofContainer& recvDofs = meshDistributor->getRecvDofs();
    for (RankToDofContainer::iterator rankIt = recvDofs.begin(); 
	 rankIt != recvDofs.end(); ++rankIt)
      for (DofContainer::iterator dofIt = rankIt->second.begin(); 
	   dofIt != rankIt->second.end(); ++dofIt)
	otherBoundaryLocalDofs.insert(**dofIt);
      
    std::set<DegreeOfFreedom> interiorDofsSet;
    DofIndexToBool& isRankDof = meshDistributor->getIsRankDof();
    for (DofIndexToBool::iterator dofIt = isRankDof.begin(); 
	 dofIt != isRankDof.end(); ++dofIt)
      if (dofIt->second && 
	  boundaryLocalDofs.count(dofIt->first) == 0 && 
	  otherBoundaryLocalDofs.count(dofIt->first) == 0)
	for (int i = 0; i < nComponents; i++)	    
	  interiorDofsSet.insert(meshDistributor->mapLocalToGlobal(dofIt->first) * nComponents + i);
      
    vector<DegreeOfFreedom> interiorDofs(interiorDofsSet.begin(), 
					 interiorDofsSet.end());

    KSPCreate(PETSC_COMM_WORLD, &solver);
    KSPGetPC(solver, &pc);
    PCSetType(pc, PCFIELDSPLIT);
      
    IS interiorIs;
    ISCreateGeneral(PETSC_COMM_WORLD, interiorDofs.size(), &(interiorDofs[0]), 
		    PETSC_COPY_VALUES, &interiorIs);
    PCFieldSplitSetIS(pc, "interior", interiorIs);
      
    IS boundaryIs;
    ISCreateGeneral(PETSC_COMM_WORLD, boundaryDofs.size(), &(boundaryDofs[0]), 
		    PETSC_COPY_VALUES, &boundaryIs);
    PCFieldSplitSetIS(pc, "boundary", boundaryIs);
     
     
    KSPSetOperators(solver, matrix, matrix, SAME_NONZERO_PATTERN); 
    KSPSetTolerances(solver, 0.0, 1e-8, PETSC_DEFAULT, PETSC_DEFAULT);
    KSPSetFromOptions(solver);
    PCSetFromOptions(pc);
  }

}
